/**********************************************************************
*                                                                     *
* Copyright (c) David L. Mills 1992-2001                              *
*                                                                     *
* Permission to use, copy, modify, and distribute this software and   *
* its documentation for any purpose and without fee is hereby         *
* granted, provided that the above copyright notice appears in all    *
* copies and that both the copyright notice and this permission       *
* notice appear in supporting documentation, and that the name        *
* University of Delaware not be used in advertising or publicity      *
* pertaining to distribution of the software without specific,        *
* written prior permission. The University of Delaware makes no       *
* representations about the suitability this software for any         *
* purpose. It is provided "as is" without express or implied          *
* warranty.                                                           *
*                                                                     *
*ChangLog							      *
*	Tony Mu:					              *		
*	fntp.c extract from ntp-4.1.71 package			      *
**********************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <utmp.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include "fntp.h"
#include "fmacro.h"
#include "ntp.h"
#include "rtc.h"
#include "SYS_log.h"
#include <arpa/inet.h>

#include "PRO_file.h"
#include "conf_sec.h"
#include "sc_tool.h"
#include "fl_conf.h"

//#define _MY_DEBUG_
#ifdef _MY_DEBUG_
#define MYPRINTF(format, argument...) printf(format , ## argument);
#else
#define MYPRINTF(format, argument...)
#endif 


int syncflag = 0;		     /* 0:Don't sync ,1:sync */
struct timeval ntp_timetv = {0,0};
int sys_numservers = 0; 	     /* number of servers to poll */
int complete_servers = 0;
int initializing = 1;
u_long current_time = 0;
volatile int alarm_flag = 0;
int sys_samples = DEFSAMPLES;	/* number of samples/server */
u_long sys_timeout = DEFTIMEOUT; /* timeout time, in TIMER_HZ units */
struct server *sys_servers;	/* the server list */
int sys_authenticate = 0;	/* true when authenticating */
u_int32 sys_authkey = 0;	/* set to authentication key in use */
int sys_version = NTP_VERSION;	/* version to poll with */
double sys_residual = 0;	/* residual from previous adjustment */


u_long volatile full_recvbufs;	     /* number of recvbufs on fulllist */
struct recvbuf *volatile freelist;   /* free buffers */
struct recvbuf *volatile fulllist;   /* lifo buffers with data */
struct recvbuf *volatile beginlist;  /* fifo buffers with data */
u_long volatile total_recvbufs;	     /* total recvbufs currently in use */
u_long volatile free_recvbufs;	     /* number of recvbufs on freelist */
u_long volatile lowater_adds;	     /* number of times we have added memory */

int fd,tzone,dzone,dsflag;
int daylight_saving=0;
fd_set fdmask;

#ifdef _USER_SETNTP_
struct timeval timeout = {20,0};
char ntpaddr[16];
#else
struct timeval timeout = {15,0};
char* server[4]={"133.100.11.8","133.100.9.2","133.88.200.22",NULL};
#endif

#if 0
#ifdef _SUPPORT_NEW_TIMEZONE_VENEZUELA_
#define TZ_NO	76
#else
#define TZ_NO	75
#endif
struct TZ_S{
	char str[80];
	int  num;
};

static struct TZ_S TimeZone[TZ_NO]={
{"(GMT-12:00) International Date Line West",-1200},
{"(GMT-11:00) Midway Island, Samoa",-1100},
{"(GMT-10:00) Hawaii",-1000},
{"(GMT-09:00) Alaska",-900},
{"(GMT-08:00) Pacific Time (US &amp; Canada); Tijuana",-800}, //default tz
{"(GMT-07:00) Arizona",-700},
{"(GMT-07:00) Chihuahua, La Paz, Mazatlan",-700},
{"(GMT-07:00) Mountain Time (US &amp; Canada)",-700},
{"(GMT-06:00) Central America",-600},
{"(GMT-06:00) Central Time (US &amp; Canda)",-600},
{"(GMT-06:00) Guadalajara, Mexico City,Monterrey",-600},
{"(GMT-06:00) Saskatchewan",-600},
{"(GMT-05:00) Bogota, Lima, Quito",-500},
{"(GMT-05:00) Eastern Time (US &amp; Canda)",-500},
{"(GMT-05:00) Indiana (East)",-500},
{"(GMT-04:00) Atlantic Time (Canada)",-400},
#ifdef _SUPPORT_NEW_TIMEZONE_VENEZUELA_
{"(GMT-04:00) La Paz",-400},
#else
{"(GMT-04:00) Caracas, La Paz",-400},
#endif
{"(GMT-04:00) Santiago",-400},
{"(GMT-03:30) Newfoundland",-350},
{"(GMT-03:00) Brasilia",-300},
{"(GMT-03:00) Buenos Aires, Georgetown",-300},
{"(GMT-03:00) Greenland",-300},
{"(GMT-02:00) Mid-Atlantic",-200},
{"(GMT-01:00) Azores",-100},
{"(GMT-01:00) Cape Verde Is.",-100},
{"(GMT) Casablanca, Monrovia",0},
{"(GMT) Greenwich Mean Time: Dublin,Edinburgh,Lisbon,London",0},
{"(GMT+01:00) Amsterdam, Berlin,Bern,Rome,Stockholm,Vienna",100},
{"(GMT+01:00) Belgrade, Bratislava,Budapest,Ljubljana,Prague",100},
{"(GMT+01:00) Brussels, Copenhagen, Madrid, Paris",100},
{"(GMT+01:00) Sarajevo, Skopje, Warsaw, Zagreb",100},
{"(GMT+01:00) West Central Africa",100},
{"(GMT+02:00) Athens, Istanbul, Minsk",200},
{"(GMT+02:00) Bucharest",200},
{"(GMT+02:00) Cairo",200},
{"(GMT+02:00) Harare, Pretoria",200},
{"(GMT+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius",200},
{"(GMT+02:00) Jerusalem",200},
{"(GMT+03:00) Baghdad",300},
{"(GMT+03:00) Kuwait, Riyadh",300},
{"(GMT+03:00) Moscow, St. Petersburg, Volgograd",300},
{"(GMT+03:00) Nairobi",300},
{"(GMT+03:30) Tehran",350},
{"(GMT+04:00) Abu Dhabi, Muscat",400},
{"(GMT+04:00) Baku, Tbilisi, Yerevan",400},
{"(GMT+04:30) Kabul",450},
{"(GMT+05:00) Ekaterinburg",500},
{"(GMT+05:00) Islamabad, Karachi, Tashkent",500},
{"(GMT+05:30) Chennai, Kolkata, Mumbai, New Delhi",550},
{"(GMT+05:45) Kathmandu",575},
{"(GMT+06:00) Almaty, Novosibirsk",600},
{"(GMT+06:00) Astana, Dhaka",600},
{"(GMT+06:00) Sri Jayawardenepura",600},
{"(GMT+06:30) Rangoon",650},
{"(GMT+07:00) Bangkok, Hanoi, Jakarta",700},
{"(GMT+07:00) Krasnoyarsk",700},
{"(GMT+08:00) Beijing, Chongqing, Hong Kong, Urumqi",800},
{"(GMT+08:00) Irkutsk, Ulaan Bataar",800},
{"(GMT+08:00) Kuala Lumpur, Singapore",800},
{"(GMT+08:00) Perth",800},
{"(GMT+08:00) Taipei",800},
{"(GMT+09:00) Osaka, Sapporo, Tokyo",900},
{"(GMT+09:00) Seoul",900},
{"(GMT+09:00) Yakutsk",900},
{"(GMT+09:30) Adelaide",950},
{"(GMT+09:30) Darwin",950},
{"(GMT+10:00) Brisbane",1000},
{"(GMT+10:00) Canberra, Melbourne, Sydney",1000},
{"(GMT+10:00) Guam, Port Moresby",1000},
{"(GMT+10:00) Hobart",1000},
{"(GMT+10:00) Vladivostok",1000},
{"(GMT+11:00) Magadan, Solomon Is., New Caledonia",1100},
{"(GMT+12:00) Auckland, Wellington",1200},
{"(GMT+12:00) Fiji, Kamchatka, Marshall Is.",1200},
{"(GMT+13:00) Nuku'alofa",1300},
#ifdef _SUPPORT_NEW_TIMEZONE_VENEZUELA_
{"(GMT-04:30) Caracas", -450},
#endif
};

#endif 

#if 0  
static int NPRO_2sect(char *sect, FILE *fp)
{
   char buf[MAX_LEN],sec[MAX_KEY_LEN],*p;
   int  ln;

   p = strstr(sect,"\n");
   if (p)
      *p = 0;
   if (sect[0] != '[')
   {
      sprintf(sec,"[%s]",sect);
      p = sec;
   }
   else
      p = sect;
   rewind(fp);
   ln = strlen(p);
   while(fgets(buf,sizeof(buf),fp))
   {
      if (buf[0] != '[')
         continue;
      if (strncmp(buf,p,ln)==0)
         return 0;
   }
   return 1;
}



int NPRO_GetStr(char *sect, char *key, char *val, int size, FILE* fp)
{
   int  i,bl;
   char *buf,*pf=NULL,*p,*p0;
   char b[MAX_LEN]; /* a line read from ini file */

   if (val==NULL || fp==NULL || key==NULL)
      return -1;

   if (sect) {
      if (NPRO_2sect(sect,fp) != 0) // section not found
         return -1;
   } else
      rewind(fp);

   if (size > MAX_LEN) {
      bl = size+MAX_KEY_LEN;
      pf = buf = malloc(bl);
      if (!buf)
         return -1;
   } else {
      bl = sizeof(b);
      buf = b;   
   }

   i = strlen(key);
   while(fgets(buf,bl,fp))
   {
      if (buf[0]=='[')
         break; // end of section
      p=buf;
      while(*p==' ')
         p++;
      if (strncmp(p,key,i)==0)
      {
         p0=p+i;
         while (*p0==' ')
            p0++;
         if (*p0 != '=')
            continue;
         p0++;
         while (*p0==' ')
            p0++;
         p=buf+strlen(buf)-1;
         while(*p==0x0d || *p==0x0a)
            p--;
         *(p+1)=0;
         if (strlen(p0) > size)
            *(p0+size-1)=0;
         strcpy(val,p0);
	 free(pf);
         return 0;
      } //if (strncmp()
   } //while (fgets())
   free(pf);
   return -1;
}

int NPRO_GetInt(char *sect, char *key, int *val, FILE *fp)
{
   char buf[MAX_KEY_LEN];
   int ret;

   ret = NPRO_GetStr(sect,key,buf,sizeof(buf),fp);
   if (ret==0)
      *val = atoi(buf);

   return ret;
}
#endif

static int get_timezone(void){
	int ret;
	FILE *fd1;

	fd1 = fopen(SYSTEM_CONF,"rt");
	if(fd1==NULL){
		return -1;
        }

	ret = PRO_GetInt(SEC_SYS, SYS_TIME_ZONE, &tzone ,fd1);	
	if (ret != 0) {
		fclose(fd1);
		return -1;
	}

	fclose(fd1);

	MYPRINTF("Get timezon index %d\n",tzone);

	return 0;
}

static int getdaylightzone(void){
	int ret;
	FILE *fd;

	fd = fopen(SYSTEM_CONF,"rt");
	if(fd==NULL){
		return -1;
        }

	ret = PRO_GetInt(SEC_SYS, SYS_DAY_SAVING, &dsflag ,fd);	

	if (ret != 0) {
		fclose(fd);
		return -1;
	}

	ret = PRO_GetInt(SEC_SYS, SYS_TIME_ZONE, &dzone ,fd);	

	if (ret != 0) {
		fclose(fd);
		return -1;
	}

	fclose(fd);

	MYPRINTF("Get day_saving enable/disable %d\n",dsflag);
	MYPRINTF("Get dzone %d\n",dzone);
	
	if (dsflag != 1)
		return 0;

	if ( (dzone==3) || (dzone==4) || (dzone==6) || (dzone==7) || (dzone==9)	|| (dzone==10) || (dzone==13) || (dzone==15) || (dzone==17) || (dzone==18) || (dzone==19) || (dzone==21) || (dzone==22) || (dzone==23) || (dzone==26) || (dzone==27) || (dzone==28) || (dzone==29) || (dzone==30) || (dzone==32) || (dzone==33) || (dzone=34) || (dzone==36) || (dzone==38) || (dzone==40) || (dzone==42) || (dzone==44) || (dzone==46) || (dzone==50) || (dzone==55) || (dzone==57) || (dzone==63) || (dzone==64) || (dzone==67) || (dzone==69) || (dzone==70) || (dzone==72)){
		
		daylight_saving=1;
	}
	return 0;
}

static int convert(int range,u8 *hi,u8 *low){
	
	*hi=range/10;
	*low=range%10;

	return 0;
}

int SETRTCTIME(void){

        time_t now;
	struct tm *now_tm;
	
	u8 rtc[16];
	u8 hi_byte,low_byte;
	

	time(&now);
	now_tm=localtime(&now);

	if(((now_tm->tm_year+1900) < 2000) || ((now_tm->tm_year+1900) >= 2038))
		return -1;

	rtc[0]=2; /* year */
	rtc[1]=0;
	rtc[2]=0; /* week */
	rtc[3]=0;

	convert((now_tm->tm_year),&hi_byte,&low_byte);
	rtc[4]=hi_byte;
	rtc[5]=low_byte;

	convert((now_tm->tm_mon),&hi_byte,&low_byte);
	rtc[6]=hi_byte;
	rtc[7]=low_byte;

	convert((now_tm->tm_mday),&hi_byte,&low_byte);
	rtc[8]=hi_byte;
	rtc[9]=low_byte;

	convert((now_tm->tm_hour),&hi_byte,&low_byte);
	rtc[10]=hi_byte;
	rtc[11]=low_byte;

	convert((now_tm->tm_min),&hi_byte,&low_byte);
	rtc[12]=hi_byte;
	rtc[13]=low_byte;

	convert((now_tm->tm_sec),&hi_byte,&low_byte);
	rtc[14]=hi_byte;
	rtc[15]=low_byte;

//printf(" The RTC date %x%x%x%x / %x%x / %x%x , week %x%x, and RTC time %x%x:%x%x:%x%x\n",rtc[0],rtc[1],rtc[4],rtc[5],rtc[6],rtc[7],rtc[8],rtc[9],rtc[2],rtc[3],rtc[10],rtc[11],rtc[12],rtc[13],rtc[14],rtc[15]);

	set_RTC_time(rtc);
	
				
	return 0;
	
}

/** Level 2 **/
static void * emalloc(u_int size)
{
	char *mem;

	if ((mem = (char *)malloc(size)) == 0) {
//		msyslog(LOG_ERR, "Exiting: No more memory!");
		exit(1);
	}
	return mem;
}

static void initialise_buffer(struct recvbuf *buff)
{
	memset((char *) buff, 0, sizeof(struct recvbuf));

}

/*
 * alarming - record the occurance of an alarm interrupt
 */

static void alarming(int sig)
{
	alarm_flag++;
}

static void signal_no_reset(int sig,void (*func)(int))
{
	int n;
	struct sigaction vec;

	vec.sa_handler = func;
	sigemptyset(&vec.sa_mask);

	vec.sa_flags = 0;

	while (1)
	{
		struct sigaction ovec;

		n = sigaction(sig, &vec, &ovec);
		if (n == -1 && errno == EINTR) continue;
		if (ovec.sa_flags)
//			msyslog(LOG_DEBUG, "signal_no_reset: signal %d had flags %x",sig, ovec.sa_flags);
;
		break;
	}
	if (n == -1) {
		perror("sigaction");
		exit(1);
	}
}

static void init_recvbuff(int nbufs)
{
	register struct recvbuf *buf;
	int i;

	/*
	 * Init buffer free list and stat counters
	 */
	freelist = 0;

	buf = (struct recvbuf *)
	    emalloc(nbufs*sizeof(struct recvbuf));
	for (i = 0; i < nbufs; i++)
	{
		initialise_buffer(buf);
		buf->next = (struct recvbuf *) freelist;
		freelist = buf;
		buf++;
	}

	fulllist = 0;
	free_recvbufs = total_recvbufs = nbufs;
	full_recvbufs = lowater_adds = 0;
}

static void create_buffers(void)
{
	register struct recvbuf *buf;
	int i;
	buf = (struct recvbuf *)
	    emalloc(RECV_INC*sizeof(struct recvbuf));
	for (i = 0; i < RECV_INC; i++)
	{
		initialise_buffer(buf);
		buf->next = (struct recvbuf *) freelist;
		freelist = buf;
		buf++;
	}

	free_recvbufs += RECV_INC;
	total_recvbufs += RECV_INC;
	lowater_adds++;
}

static int decodenetnum(const char *num, u_int32 *netnum)
{
	register const char *cp;
	register char *bp;
	register int i;
	register int temp;
	register int eos;
	char buf[80];		/* will core dump on really stupid stuff */

	cp = num;
	*netnum = 0;

	if (*cp == '[') {
		eos = ']';
		cp++;
	} else {
		eos = '\0';
	}

	for (i = 0; i < 4; i++) {
		bp = buf;
		while (isdigit((int)*cp))
		    *bp++ = *cp++;
		if (bp == buf)
		    break;

		if (i < 3) {
			if (*cp++ != '.')
			    break;
		} else if (*cp != eos)
		    break;

		*bp = '\0';
		temp = atoi(buf);
		if (temp > 255)
		    break;
		*netnum <<= 8;
		*netnum += temp;
	}
	
	if (i < 4)
	    return 0;
	*netnum = htonl(*netnum);
	return 1;
}


/*
 * addserver - determine a server's address and allocate a new structure
 *		   for it.
 */

static void addserver(char *serv)
{
	register struct server *server;
	u_int32 netnum;

	decodenetnum(serv, &netnum);
	server = (struct server *)emalloc(sizeof(struct server));
	memset((char *)server, 0, sizeof(struct server));

	server->srcadr.sin_family = AF_INET;
	server->srcadr.sin_addr.s_addr = netnum;
	server->srcadr.sin_port = htons(NTP_PORT);

	server->event_time = ++sys_numservers;
	if (sys_servers == NULL){
		sys_servers = server;
	}
	else {
		struct server *sp;

		for (sp = sys_servers; sp->next_server != NULL;
		     sp = sp->next_server) ;
		sp->next_server = server;
	}
}



/** Level 1 **/
/*
void msyslog(int level, const char *fmt, ...)
{
	va_list ap;
	char buf[1025], nfmt[256];
	register int c;
	register char *n, *prog;
	register const char *f;
	int olderrno;
	char *err;

	va_start(ap, fmt);

	olderrno = errno;
	n = nfmt;
	f = fmt;
	while ((c = *f++) != '\0' && c != '\n' && n < &nfmt[252]) {
		if (c != '%') {
			*n++ = c;
			continue;
		}
		if ((c = *f++) != 'm') {
			*n++ = '%';
			*n++ = c;
			continue;
		}
		err = 0;

		err = strerror(olderrno);

		if ((n + strlen(err)) < &nfmt[254]) {
			strcpy(n, err);
			n += strlen(err);
		}
	}
	if (!syslogit)
	    *n++ = '\n';
	*n = '\0';

	vsprintf(buf, nfmt, ap);
	if (syslogit)
	    syslog(level, "%s", buf);
	else
	{
		FILE *out_file = syslog_file ? syslog_file
			: level <= LOG_ERR ? stderr : stdout;
		prog = strrchr(progname, '/');
		if (prog == NULL)
		    prog = progname;
		else
		    prog++;
		(void) fprintf(out_file, "%s ", humanlogtime ());
		(void) fprintf(out_file, "%s[%d]: %s", prog, (int)getpid(), buf);
		fflush (out_file);
	}
	va_end(ap);
}
*/

/*
 * init_io - initialize I/O data and open socket
 */
static void init_io(void)
{
	int on = 0;
	struct sockaddr_in addr;

	/*
	 * Init buffer free list and stat counters
	 */
	init_recvbuff(sys_numservers + 2);
	/*
	 * Open the socket
	 */

	/* create a datagram (UDP) socket */
	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
//		msyslog(LOG_ERR, "socket() failed: %m");
		exit(1);
		/*NOTREACHED*/
	}
	setsockopt(fd, SOL_IP, IP_MTU_DISCOVER, (char*)&on, sizeof(on));

	/*
	 * bind the socket to the NTP port
	 */

		memset((char *)&addr, 0, sizeof addr);
		addr.sin_family = AF_INET;
		addr.sin_port = htons(NTP_PORT);
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
		if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
			exit(1);
		}

	FD_ZERO(&fdmask);
	FD_SET(fd, &fdmask);

	/*
	 * set non-blocking,
	 */
	if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0) {
		exit(1);
		/*NOTREACHED*/
	}
}


static void init_alarm(void)
{
	struct itimerval itimer;

	alarm_flag = 0;

	/*
	 * Set up the alarm interrupt.	The first comes 1/(2*TIMER_HZ)
	 * seconds from now and they continue on every 1/TIMER_HZ seconds.
	 */
	(void) signal_no_reset(SIGALRM,alarming);
	itimer.it_interval.tv_sec = itimer.it_value.tv_sec = 0;
	itimer.it_interval.tv_usec = 1000000/TIMER_HZ;
	itimer.it_value.tv_usec = 1000000/(TIMER_HZ<<1);
	setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);
}

/*
 * getrecvbufs - get receive buffers which have data in them
 */

struct recvbuf * getrecvbufs(void){
	struct recvbuf *rb = NULL; /* nothing has arrived */;

	if (full_recvbufs == 0)
	{
		;
	}
	else {

		/*
		 * Get the fulllist chain and mark it empty
		 */
		rb = beginlist;
		fulllist = 0;
		full_recvbufs = 0;

		/*
		 * Check to see if we're below the low water mark.
		 */
		if (free_recvbufs <= RECV_LOWAT)
		{
			if (total_recvbufs >= RECV_TOOMANY)
//			    msyslog(LOG_ERR, "too many recvbufs allocated (%ld)",
//				    total_recvbufs);
;
			else
			{
				create_buffers();
			}
		}
	}

	/*
	 * Return the chain
	 */
	return rb;
}

/*
 * server_data - add a sample to the server's filter registers
 */
static void server_data(register struct server *server,s_fp d,l_fp *c,u_fp e)
{
	register int i;

	i = server->filter_nextpt;
	if (i < NTP_SHIFT) {
		server->filter_delay[i] = d;
		server->filter_offset[i] = *c;
		server->filter_soffset[i] = LFPTOFP(c);
		server->filter_error[i] = e;
		server->filter_nextpt = i + 1;
	}
}

/*
 * get_systime - return the system time in timestamp format biased by
 * the current time offset.
 */
void get_systime(l_fp *now)
{
	struct timeval tv;
	double dtemp;
	int *TimeZone = get_time_zone_table();
	/*
	 * We use nanosecond time if we can get it. Watch out for
	 * rounding wiggles, which may overflow the fraction.
	 */
	(void) GETTIMEOFDAY(&tv, (struct timezone *)0);
	//tv.tv_sec -= (tzone*3600);
	tv.tv_sec -= (TimeZone[tzone]*36);
	now->l_i = tv.tv_sec + JAN_1970;


	dtemp = tv.tv_usec * FRAC / 1e6;

	if (dtemp >= FRAC)
		now->l_i++;
	now->l_uf = (u_int32)dtemp;

}

/*
 * sendpkt - send a packet to the specified destination
 */
static void sendpkt(struct sockaddr_in *dest, struct pkt *pkt,int len)
{
	int cc;

	cc = sendto(fd, (char *)pkt, (size_t)len, 0, (struct sockaddr *)dest,
			sizeof(struct sockaddr_in));
	if (cc == -1) {
		if (errno != EWOULDBLOCK && errno != ENOBUFS)
//			msyslog(LOG_ERR, "sendto(%s): %m", ntoa(dest));
;
	}
}


/*
 * transmit - transmit a packet to the given server, or mark it completed.
 *  This is called by the timeout routine and by the receive
 *		  procedure.
 */
static void transmit(register struct server *server)
{
	struct pkt xpkt;

	if (server->filter_nextpt < server->xmtcnt) {
		l_fp ts;
		/*
		 * Last message to this server timed out.  Shift
		 * zeros into the filter.
		 */
		L_CLR(&ts);
		server_data(server, 0, &ts, 0);
	}

	if ((int)server->filter_nextpt >= sys_samples) {
		/*
		 * Got all the data we need.  Mark this guy
		 * completed and return.
		 */
		server->event_time = 0;
		complete_servers++;
		return;
	}

	/*
	 * If we're here, send another message to the server.    Fill in
	 * the packet and let 'er rip.
	 */
	xpkt.li_vn_mode = PKT_LI_VN_MODE(LEAP_NOTINSYNC,
					 sys_version, MODE_CLIENT);
	xpkt.stratum = STRATUM_TO_PKT(STRATUM_UNSPEC);
	xpkt.ppoll = NTP_MINPOLL;
	xpkt.precision = NTPDATE_PRECISION;
	xpkt.rootdelay = htonl(NTPDATE_DISTANCE);
	xpkt.rootdispersion = htonl(NTPDATE_DISP);
	xpkt.refid = htonl(NTPDATE_REFID);
	L_CLR(&xpkt.reftime);
	L_CLR(&xpkt.org);
	L_CLR(&xpkt.rec);

	/*
	 * Determine whether to authenticate or not.	If so,
	 * fill in the extended part of the packet and do it.
	 * If not, just timestamp it and send it away.
	 */
		get_systime(&(server->xmt));
		HTONL_FP(&server->xmt, &xpkt.xmt);
		sendpkt(&(server->srcadr), &xpkt, LEN_PKT_NOMAC);


	/*
	 * Update the server timeout and transmit count
	 */
	server->event_time = current_time + sys_timeout;
	server->xmtcnt++;
}


/*
 * findserver - find a server in the list given its address
 */
static struct server *findserver(struct sockaddr_in *addr)
{
	register u_int32 netnum;
	struct server *server;
	struct server *mc_server;

	mc_server = NULL;
	if (htons(addr->sin_port) != NTP_PORT)
		return 0;
	netnum = addr->sin_addr.s_addr;

	for (server = sys_servers; server != NULL; 
	     server = server->next_server) {
		register u_int32 servnum;
		
		servnum = server->srcadr.sin_addr.s_addr;
		if (netnum == servnum)
			return server;
		if (IN_MULTICAST(ntohl(servnum)))
			mc_server = server;
	}

	if (mc_server != NULL) {	
		struct server *sp;

		if (mc_server->event_time != 0) {
			mc_server->event_time = 0;
			complete_servers++;
		}
		server = (struct server *)emalloc(sizeof(struct server));
		memset((char *)server, 0, sizeof(struct server));

		server->srcadr.sin_family = AF_INET;
		server->srcadr.sin_addr.s_addr = netnum;
		server->srcadr.sin_port = htons(NTP_PORT);

		server->event_time = ++sys_numservers;
		for (sp = sys_servers; sp->next_server != NULL;
		     sp = sp->next_server) ;
		sp->next_server = server;
		transmit(server);
	}
	return NULL;
}

u_long authdecryptions;		/* calls to decrypt */
/*
 * authdecrypt - verify message authenticator
 *
 * Returns one if authenticator valid, zero if invalid or key not found.
 */
static int authdecrypt(keyid_t keyno,u_int32 *pkt,int length,int size)
{
	/*
	 * A zero key identifier means the sender has not verified
	 * the last message was correctly authenticated. Nevertheless,
	 * the authenticator itself is considered valid.
	 */
	authdecryptions++;
	if (keyno == 0)
		return (0);

	return (0);
}

/*
 * receive - receive and process an incoming frame
 */
static void receive(struct recvbuf *rbufp)
{
	register struct pkt *rpkt;
	register struct server *server;
	register s_fp di;
	l_fp t10, t23;
	l_fp org;
	l_fp rec;
	l_fp ci;
	int has_mac;
	int is_authentic;

	/*
	 * Check to see if the packet basically looks like something
	 * intended for us.
	 */
	if (rbufp->recv_length == LEN_PKT_NOMAC)
		has_mac = 0;
	else if (rbufp->recv_length >= LEN_PKT_NOMAC)
		has_mac = 1;
	else {
		return; 		/* funny length packet */
	}

	rpkt = &(rbufp->recv_pkt);
	if (PKT_VERSION(rpkt->li_vn_mode) < NTP_OLDVERSION ||
		PKT_VERSION(rpkt->li_vn_mode) > NTP_VERSION) {
		return;
	}

	if ((PKT_MODE(rpkt->li_vn_mode) != MODE_SERVER
		 && PKT_MODE(rpkt->li_vn_mode) != MODE_PASSIVE)
		|| rpkt->stratum >= STRATUM_UNSPEC) {
		return;
	}

	/*
	 * So far, so good.  See if this is from a server we know.
	 */
	server = findserver(&(rbufp->recv_srcadr));
	if (server == NULL) {
		return;
	}

	/*
	 * Decode the org timestamp and make sure we're getting a response
	 * to our last request.
	 */
	NTOHL_FP(&rpkt->org, &org);
	if (!L_ISEQU(&org, &server->xmt)) {
		return;
	}

	/*
	 * Check out the authenticity if we're doing that.
	 */
	if (!sys_authenticate)
		is_authentic = 1;
	else {
		is_authentic = 0;


		if (has_mac && ntohl(rpkt->exten[0]) == sys_authkey &&
			authdecrypt(sys_authkey, (u_int32 *)rpkt, LEN_PKT_NOMAC,
			(int)(rbufp->recv_length - LEN_PKT_NOMAC)))
			is_authentic = 1;
	}
	server->trust <<= 1;
	if (!is_authentic)
		server->trust |= 1;

	/*
	 * Looks good.	Record info from the packet.
	 */
	server->leap = PKT_LEAP(rpkt->li_vn_mode);
	server->stratum = PKT_TO_STRATUM(rpkt->stratum);
	server->precision = rpkt->precision;
	server->rootdelay = ntohl(rpkt->rootdelay);
	server->rootdispersion = ntohl(rpkt->rootdispersion);
	server->refid = rpkt->refid;
	NTOHL_FP(&rpkt->reftime, &server->reftime);
	NTOHL_FP(&rpkt->rec, &rec);
	NTOHL_FP(&rpkt->xmt, &server->org);

	/*
	 * Make sure the server is at least somewhat sane.	If not, try
	 * again.
	 */
	if (L_ISZERO(&rec) || !L_ISHIS(&server->org, &rec)) {
		transmit(server);
		return;
	}

	/*
	 * Calculate the round trip delay (di) and the clock offset (ci).
	 * We use the equations (reordered from those in the spec):
	 *
	 * d = (t2 - t3) - (t1 - t0)
	 * c = ((t2 - t3) + (t1 - t0)) / 2
	 */
	t10 = server->org;		/* pkt.xmt == t1 */
	L_SUB(&t10, &rbufp->recv_time); /* recv_time == t0*/

	t23 = rec;			/* pkt.rec == t2 */
	L_SUB(&t23, &org);		/* pkt->org == t3 */

	/* now have (t2 - t3) and (t0 - t1).	Calculate (ci) and (di) */
	ci = t10;
	L_ADD(&ci, &t23);
	L_RSHIFT(&ci);

	/*
	 * Calculate di in t23 in full precision, then truncate
	 * to an s_fp.
	 */
	L_SUB(&t23, &t10);
	di = LFPTOFP(&t23);

	di += (FP_SECOND >> (-(int)NTPDATE_PRECISION))
		+ (FP_SECOND >> (-(int)server->precision)) + NTP_MAXSKW;

	if (di <= 0) {		/* value still too raunchy to use? */
		L_CLR(&ci);
		di = 0;
	} else {
		di = max(di, NTP_MINDIST);
	}

	/*
	 * Shift this data in, then transmit again.
	 */
	server_data(server, (s_fp) di, &ci, 0);
	transmit(server);
}




/*
 * freerecvbuf - make a single recvbuf available for reuse
 */
static void freerecvbuf(struct recvbuf *rb)
{
	rb->next = (struct recvbuf *) freelist;
	freelist = rb;
	free_recvbufs++;
}

/*
 * timer - process a timer interrupt
 */
static void timer(void){
	struct server *server;

	/*
	 * Bump the current idea of the time
	 */
	current_time++;

	/*
	 * Search through the server list looking for guys
	 * who's event timers have expired.  Give these to
	 * the transmit routine.
	 */
	for (server = sys_servers; server != NULL; 
	     server = server->next_server) {
		if (server->event_time != 0
		    && server->event_time <= current_time)
			transmit(server);
	}
}



const char *set_tod_using = "UNKNOWN";
static int ntp_set_tod(struct timeval *tvp,void *tzp)
{
	int rc;
	int *TimeZone = get_time_zone_table();
#ifndef _SC_DST_CHANGE_AUTO_	
		if (daylight_saving == 1)
			tvp->tv_sec += 36;
#endif

//		tvp->tv_sec += (tzone*3600);
		tvp->tv_sec += (TimeZone[tzone]*36);

		if(syncflag)
		{
			rc = SETTIMEOFDAY(tvp, tzp);
			SETRTCTIME();
			if (!rc)
			{
				set_tod_using = "settimeofday";
			#ifdef _SC_DST_CHANGE_AUTO_
				/*
				 * To avoid the syscall() interrupt by the signal SIGALRM when was used at get 
				 * NTP time from NTP Server, we should reset the signal.
				 * To avoid time in log not correct, ntpdate need wait until syscall(sc_dst_setcron)  return
				 */
				signal(SIGALRM, SIG_IGN);   
				PRO_SetInt(SEC_MAN, MAN_NTP_SUMMER_CHG, 0, SYSTEM_CONF);
				fl_write_conf(SYSTEM_CONF);
				//system("/usr/local/bin/sc_dst_setcron 1 0  >/dev/null 2>/dev/null ");			
			#endif							
				return rc;
			}
		
			set_tod_using = "Failed!";
			return -1;
		}
		
		ntp_timetv.tv_sec=tvp->tv_sec;
		ntp_timetv.tv_usec=tvp->tv_usec;

	return 0;
}

/*
 * step_systime - step the system clock.
 */
static int step_systime(double now)
{
	struct timeval timetv, adjtv, oldtimetv;
	int isneg = 0;
	double dtemp;
	int *TimeZone = get_time_zone_table();
	dtemp = sys_residual + now;
	if (dtemp < 0) {
		isneg = 1;
		dtemp = - dtemp;
		adjtv.tv_sec = (int32)dtemp;
		adjtv.tv_usec = (u_int32)((dtemp - (double)adjtv.tv_sec) *
					  1e6 + .5);
	} else {
		adjtv.tv_sec = (int32)dtemp;
		adjtv.tv_usec = (u_int32)((dtemp - (double)adjtv.tv_sec) *
					  1e6 + .5);
	}
	(void) GETTIMEOFDAY(&timetv, (struct timezone *)0);
//	timetv.tv_sec -= (tzone*3600);
	timetv.tv_sec -= (TimeZone[tzone]*36);

	oldtimetv = timetv;

	if (isneg) {
		timetv.tv_sec -= adjtv.tv_sec;
		timetv.tv_usec -= adjtv.tv_usec;
		if (timetv.tv_usec < 0) {
			timetv.tv_sec--;
			timetv.tv_usec += 1000000;
		}
	} else {
		timetv.tv_sec += adjtv.tv_sec;
		timetv.tv_usec += adjtv.tv_usec;
		if (timetv.tv_usec >= 1000000) {
			timetv.tv_sec++;
			timetv.tv_usec -= 1000000;
		}
	}
	if (ntp_set_tod(&timetv, (struct timezone *)0) != 0) {
//		msyslog(LOG_ERR, "Can't set time of day: %m");
		return (0);
	}
	sys_residual = 0;

	/*
	 * FreeBSD, for example, has:
	 * struct utmp {
	 *	   char    ut_line[UT_LINESIZE];
	 *	   char    ut_name[UT_NAMESIZE];
	 *	   char    ut_host[UT_HOSTSIZE];
	 *	   long    ut_time;
	 * };
	 * and appends line="|", name="date", host="", time for the OLD
	 * and appends line="{", name="date", host="", time for the NEW
	 * to _PATH_WTMP .
	 *
	 * Some OSes have utmp, some have utmpx.
	 */

	/*
	 * Write old and new time entries in utmp and wtmp if step adjustment
	 * is greater than one second.
	 *
	 * This might become even Uglier...
	 */
	if (oldtimetv.tv_sec != timetv.tv_sec)
	{
		struct utmp ut;

		memset((char *)&ut, 0, sizeof(ut));

	}
	return (1);
}



/*
 * This fuction is not the same as lib/systime step_systime!!!
 */
static int l_step_systime(l_fp *ts)
{
	double dtemp;

	LFPTOD(ts, dtemp);
	return step_systime(dtemp);
}


/*
 * clock_select - select the pick-of-the-litter clock from the samples
 *		  we've got.
 */
static struct server *clock_select(void)
{
	register struct server *server;
	register int i;
	register int nlist;
	register s_fp d;
	register int j;
	register int n;
	s_fp local_threshold;
	struct server *server_list[NTP_MAXCLOCK];
	u_fp server_badness[NTP_MAXCLOCK];
	struct server *sys_server;

	/*
	 * This first chunk of code is supposed to go through all
	 * servers we know about to find the NTP_MAXLIST servers which
	 * are most likely to succeed.	We run through the list
	 * doing the sanity checks and trying to insert anyone who
	 * looks okay.	We are at all times aware that we should
	 * only keep samples from the top two strata and we only need
	 * NTP_MAXLIST of them.
	 */
	nlist = 0;	/* none yet */
	for (server = sys_servers; server != NULL; server = server->next_server) {
		if (server->delay == 0)
			continue;	/* no data */
		if (server->stratum > NTP_INFIN)
			continue;	/* stratum no good */
		if (server->delay > NTP_MAXWGT) {
			continue;	/* too far away */
		}
		if (server->leap == LEAP_NOTINSYNC)
			continue;	/* he's in trouble */
		if (!L_ISHIS(&server->org, &server->reftime)) {
			continue;	/* very broken host */
		}
		if ((server->org.l_ui - server->reftime.l_ui)
			>= NTP_MAXAGE) {
			continue;	/* too long without sync */
		}
		if (server->trust != 0) {
			continue;
		}

		/*
		 * This one seems sane.  Find where he belongs
		 * on the list.
		 */
		d = server->dispersion + server->dispersion;
		for (i = 0; i < nlist; i++)
			if (server->stratum <= server_list[i]->stratum)
			break;
		for ( ; i < nlist; i++) {
			if (server->stratum < server_list[i]->stratum)
				break;
			if (d < (s_fp) server_badness[i])
				break;
		}

		/*
		 * If i points past the end of the list, this
		 * guy is a loser, else stick him in.
		 */
		if (i >= NTP_MAXLIST)
			continue;
		for (j = nlist; j > i; j--)
			if (j < NTP_MAXLIST) {
				server_list[j] = server_list[j-1];
				server_badness[j]
					= server_badness[j-1];
			}

		server_list[i] = server;
		server_badness[i] = d;
		if (nlist < NTP_MAXLIST)
			nlist++;
	}

	/*
	 * Got the five-or-less best.	 Cut the list where the number of
	 * strata exceeds two.
	 */
	j = 0;
	for (i = 1; i < nlist; i++)
		if (server_list[i]->stratum > server_list[i-1]->stratum)
		if (++j == 2) {
			nlist = i;
			break;
		}

	/*
	 * Whew!  What we should have by now is 0 to 5 candidates for
	 * the job of syncing us.  If we have none, we're out of luck.
	 * If we have one, he's a winner.  If we have more, do falseticker
	 * detection.
	 */

	if (nlist == 0)
		sys_server = 0;
	else if (nlist == 1) {
		sys_server = server_list[0];
	} else {
		/*
		 * Re-sort by stratum, bdelay estimate quality and
		 * server.delay.
		 */
		for (i = 0; i < nlist-1; i++)
			for (j = i+1; j < nlist; j++) {
				if (server_list[i]->stratum
				< server_list[j]->stratum)
				break;	/* already sorted by stratum */
				if (server_list[i]->delay
				< server_list[j]->delay)
				continue;
				server = server_list[i];
				server_list[i] = server_list[j];
				server_list[j] = server;
			}

		/*
		 * Calculate the fixed part of the dispersion limit
		 */
		local_threshold = (FP_SECOND >> (-(int)NTPDATE_PRECISION))
			+ NTP_MAXSKW;

		/*
		 * Now drop samples until we're down to one.
		 */
		while (nlist > 1) {
			for (n = 0; n < nlist; n++) {
				server_badness[n] = 0;
				for (j = 0; j < nlist; j++) {
					if (j == n) /* with self? */
						continue;
					d = server_list[j]->soffset
						- server_list[n]->soffset;
					if (d < 0)	/* absolute value */
						d = -d;
					/*
					 * XXX This code *knows* that
					 * NTP_SELECT is 3/4
					 */
					for (i = 0; i < j; i++)
						d = (d>>1) + (d>>2);
					server_badness[n] += d;
				}
			}

			/*
			 * We now have an array of nlist badness
			 * coefficients.	Find the badest.  Find
			 * the minimum precision while we're at
			 * it.
			 */
			i = 0;
			n = server_list[0]->precision;;
			for (j = 1; j < nlist; j++) {
				if (server_badness[j] >= server_badness[i])
					i = j;
				if (n > server_list[j]->precision)
					n = server_list[j]->precision;
			}

			/*
			 * i is the index of the server with the worst
			 * dispersion.	If his dispersion is less than
			 * the threshold, stop now, else delete him and
			 * continue around again.
			 */
			if ( (s_fp) server_badness[i] < (local_threshold
							 + (FP_SECOND >> (-n))))
				break;
			for (j = i + 1; j < nlist; j++)
				server_list[j-1] = server_list[j];
			nlist--;
		}

		/*
		 * What remains is a list of less than 5 servers.  Take
		 * the best.
		 */
		sys_server = server_list[0];
	}

	/*
	 * That's it.    Return our server.
	 */
	return sys_server;
}

/*
 * clock_filter - determine a server's delay, dispersion and offset
 */
static void clock_filter(register struct server *server)
{
	register int i, j;
	int ord[NTP_SHIFT];

	/*
	 * Sort indices into increasing delay order
	 */
	for (i = 0; i < sys_samples; i++)
		ord[i] = i;

	for (i = 0; i < (sys_samples-1); i++) {
		for (j = i+1; j < sys_samples; j++) {
			if (server->filter_delay[ord[j]] == 0)
				continue;
			if (server->filter_delay[ord[i]] == 0
				|| (server->filter_delay[ord[i]]
				> server->filter_delay[ord[j]])) {
				register int tmp;

				tmp = ord[i];
				ord[i] = ord[j];
				ord[j] = tmp;
			}
		}
	}

	/*
	 * Now compute the dispersion, and assign values to delay and
	 * offset.	If there are no samples in the register, delay and
	 * offset go to zero and dispersion is set to the maximum.
	 */
	if (server->filter_delay[ord[0]] == 0) {
		server->delay = 0;
		L_CLR(&server->offset);
		server->soffset = 0;
		server->dispersion = PEER_MAXDISP;
	} else {
		register s_fp d;

		server->delay = server->filter_delay[ord[0]];
		server->offset = server->filter_offset[ord[0]];
		server->soffset = LFPTOFP(&server->offset);
		server->dispersion = 0;
		for (i = 1; i < sys_samples; i++) {
			if (server->filter_delay[ord[i]] == 0)
				d = PEER_MAXDISP;
			else {
				d = server->filter_soffset[ord[i]]
					- server->filter_soffset[ord[0]];
				if (d < 0)
					d = -d;
				if (d > PEER_MAXDISP)
					d = PEER_MAXDISP;
			}
			/*
			 * XXX This *knows* PEER_FILTER is 1/2
			 */
			server->dispersion += (u_fp)(d) >> i;
		}
	}
	/*
	 * We're done
	 */
}



/*
 * clock_adjust - process what we've received, and adjust the time
 *		 if we got anything decent.
 */
static int clock_adjust(void)
{
	register struct server *sp, *server, *esp;
	register struct recvbuf *buf ,*ebuf;
	int ret=0;

	for (sp = sys_servers; sp != NULL; sp = sp->next_server)
		clock_filter(sp);
	server = clock_select();

	if (server == 0) {
		for (sp = sys_servers; sp != NULL; ){
			esp=sp;
			sp = sp->next_server;
			free(esp);
		}

		for (buf = freelist ; buf != NULL; ){
			ebuf=buf;
			buf = buf->next;
			free(ebuf);
		}

		AddLog(SYSLOG_SECTION_NTP,SYSLOG_WARNING,3);
		return(-1);
	}
	
	//printf("NTP:%s\n",inet_ntoa(server->srcadr.sin_addr));

	if (l_step_systime(&server->offset)) {
		AddLog(SYSLOG_SECTION_NTP,SYSLOG_NORMAL, 1, inet_ntoa(server->srcadr.sin_addr));
		ret=0;
	}
	else
	{
		ret =-1;
	}

	for (sp = sys_servers; sp != NULL; ){
		esp=sp;
		sp = sp->next_server;
		free(esp);
	}

	for (buf = freelist ; buf != NULL; ){
		ebuf=buf;
		buf = buf->next;
		free(ebuf);
	}

	return ret;
}

static void add_full_recv_buffer(struct recvbuf *rb)
{
	if (full_recvbufs == 0)
	{
		beginlist = rb;
		rb->next = 0;
	}
	else
	{
		rb->next = fulllist->next;
		fulllist->next = rb;
	}
	fulllist = rb;
	full_recvbufs++;
}

struct recvbuf * get_free_recv_buffer(void)
{
	struct recvbuf * buffer = NULL;
	if (free_recvbufs <= RECV_LOWAT)
		{
			if (total_recvbufs >= RECV_TOOMANY) {
			}
			else
			{
				create_buffers();
			}
		}

	if (free_recvbufs > 0)
	{
		buffer = freelist;
		freelist = buffer->next;
		buffer->next = NULL;
		--free_recvbufs;
	}
	return buffer;
}

static u_long free_recvbuffs (void)
{
	return free_recvbufs;
}


/*
 * input_handler - receive packets asynchronously
 */
static void input_handler(void)
{
	register int n;
	register struct recvbuf *rb;
	struct timeval tvzero;
	int fromlen;
	l_fp ts;
	fd_set fds;

	/*
	 * Do a poll to see if we have data
	 */
	for (;;) {
		fds = fdmask;
		tvzero.tv_sec = tvzero.tv_usec = 0;
		n = select(fd+1, &fds, (fd_set *)0, (fd_set *)0, &tvzero);

		/*
		 * If nothing to do, just return.  If an error occurred,
		 * complain and return.  If we've got some, freeze a
		 * timestamp.
		 */
		if (n == 0)
			return;
		else if (n == -1) {
			if (errno != EINTR)
			return;
		}
		get_systime(&ts);

		/*
		 * Get a buffer and read the frame.  If we
		 * haven't got a buffer, or this is received
		 * on the wild card socket, just dump the packet.
		 */
		if (initializing || free_recvbuffs() == 0) {
			char buf[100];

			(void) read(fd, buf, sizeof buf);
			continue;
		}

		rb = get_free_recv_buffer();

		fromlen = sizeof(struct sockaddr_in);
		rb->recv_length = recvfrom(fd, (char *)&rb->recv_pkt,
		   sizeof(rb->recv_pkt), 0,
		   (struct sockaddr *)&rb->recv_srcadr, &fromlen);
		if (rb->recv_length == -1) {
			freerecvbuf(rb);
			continue;
		}

		/*
		 * Got one.  Mark how and when it got here,
		 * put it on the full list.
		 */
		rb->recv_time = ts;
		add_full_recv_buffer(rb);
	}
}


int NTP_update(char *serv){
	int was_alarmed;
	struct recvbuf *rbuflist;
	struct recvbuf *rbuf;
#ifndef _USER_SETNTP_
	int i;
#endif	
	
	syncflag=1;

#ifdef _USER_SETNTP_
	addserver(serv);
#else
	for(i=0; i<3; i++)	
		addserver(server[i]);
#endif

	if (sys_numservers == 0) {
	    AddLog(SYSLOG_SECTION_NTP,SYSLOG_WARNING,4);
		return (1);
		//exit(1);
	}

	get_timezone();
	getdaylightzone();
	
	init_io();
	
	init_alarm();
	/*
	 * Set the priority.
	 */

	initializing = 0;
	was_alarmed = 0;
	rbuflist = (struct recvbuf *)0;
	

	//while (complete_servers < sys_numservers) {
	while (complete_servers < 1) {
	
		fd_set rdfdes;
		int nfound;

		if (alarm_flag) {		/* alarmed? */
			was_alarmed = 1;
			alarm_flag = 0;
		}
		rbuflist = getrecvbufs();	/* get received buffers */

		if (!was_alarmed && rbuflist == (struct recvbuf *)0) {
			/*
			 * Nothing to do.	 Wait for something.
			 */
			rdfdes = fdmask;
			nfound = select(fd+1, &rdfdes, (fd_set *)0,
					(fd_set *)0, &timeout);

			if (nfound > 0){
				input_handler();
			}
			else if (
				nfound == -1
				) {
				if (errno != EINTR)
				;
			} else {
//				AddLog(SYSLOG_ALL,"NTP select(): nfound  error");
			}
			if (alarm_flag) {		/* alarmed? */
				was_alarmed = 1;
				alarm_flag = 0;
			}
			rbuflist = getrecvbufs();	/* get received buffers */
		}

		/*
		 * Out here, signals are unblocked.  Call receive
		 * procedure for each incoming packet.
		 */
		while (rbuflist != (struct recvbuf *)0) {
			rbuf = rbuflist;
			rbuflist = rbuf->next;
			receive(rbuf);
			freerecvbuf(rbuf);
		}

		/*
		 * Call timer to process any timeouts
		 */
		if (was_alarmed) {
			timer();
			was_alarmed = 0;
		}

		/*
		 * Go around again
		 */
	}
	
	close(fd);
	
	return clock_adjust();
}
